/*
 * Copyright (C) 2011 Purdue University
 * Written by Gregor Richards
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

(function() {
    // put benchmarks here
    var benchmarks = ["jsmipsdc"];
    var modes = {
        "*": ["urv", "ure"],
        "jsmipsdc": ["urv", "ure", "ured", "usre", "usred"]
    };
    var minRuns = 25;
    var maxRuns = 1000;
    var maxSEMM = 0.05;

    // all test results
    var results = {};

    var curRun = 0;
    var curBenchmark = 0;
    var curMode = 0;

    function getModes(bm) {
        if (bm in modes) return modes[bm];
        return modes["*"];
    }

    // call this to either rerun or go to another page and come back
    function rerun() {
        try {
            if (window.sessionStorage) {
                // store this for the session and go 'round to another page to force uncaching on Chrome
                sessionStorage.JSBNG_harnessState = JSON.stringify({
                    results: results,
                    curRun: curRun,
                    curBenchmark: curBenchmark,
                    curMode: curMode
                });
                window.location.href = window.location.href.replace(/\/[^\/]*$/, "/reload.html");
                return;
            }
        } catch (ex) {}
        runBenchmark();
    }

    // load our current state and run
    function onload() {
        try {
            if (window.sessionStorage && "JSBNG_harnessState" in sessionStorage) {
                var state = JSON.parse(sessionStorage.JSBNG_harnessState);
                results = state.results;
                curRun = state.curRun;
                curBenchmark = state.curBenchmark;
                curMode = state.curMode;
            }
        } catch (ex) {}
        setTimeout(runBenchmark, 200);
    }

    function runBenchmark() {
        var output = document.getElementById("output");
        var bmframe = document.getElementById("bmframe");

        // should we stop?
        if (curRun < minRuns) {
            output.innerHTML = (curRun+1) + "/" + minRuns;
        } else {
            // check if our S/M is OK
            var stats = handleResults(false);
            output.innerHTML = (curRun+1) + " " + stats.semm + " (" + maxSEMM + ")";
            if (curRun >= maxRuns || stats.semm < maxSEMM) {
                bmframe.src = "about:blank";
                bmframe.style.display = "none";
                handleResults(true);
                if (window.sessionStorage) delete sessionStorage.JSBNG_harnessState;
                return;
            }
        }

        // get out our benchmark and mode
        var benchmark = benchmarks[curBenchmark];
        var modes = getModes(benchmark);
        var mode = modes[curMode];
        if (++curMode >= modes.length) {
            curMode = 0;
            curBenchmark++;
        }
        if (curBenchmark >= benchmarks.length) {
            curBenchmark = 0;
            curRun++;
        }

        // make sure we have the results space
        if (!(benchmark in results)) results[benchmark] = {};
        if (!(mode in results[benchmark])) results[benchmark][mode] = [];

        // set up the receiver
        if (/v/.test(mode)) {
            // verification mode, we only care if there's an error
            window.JSBNG_handleResult = function(res) {
                if (res.error) {
                    if (!("errors" in results)) results.errors = [];
                    results.errors.push(benchmark + "." + mode + ": " + res.msg);
                }
                rerun();
            };
        } else {
            window.JSBNG_handleResult = function(res) {
                if (!res.error) {
                    results[benchmark][mode].push(res.time);
                }
                rerun();
            };
        }

        // then load it
        bmframe.src = benchmark + "/" + mode + ".html";
    }

    // handle all of our results
    function handleResults(pr) {
        var output = document.getElementById("output");
        function print(str) {
            output.appendChild(document.createTextNode(str));
            output.appendChild(document.createElement("br"));
        }
        function printarr(arr) {
            for (var i = 0; i < arr.length; i++) print(arr[i]);
        }
        function percent(num) {
            return (num*100).toFixed(2) + "%";
        }

        // clear out the intermediate results
        if (pr) output.innerHTML = "";

        // totals
        var totals = {
            mean: 0,
            sm: 0,
            semm: 0,
            runs: 0
        };

        // stuff to print later
        var ptotals = [];
        var presults = [];
        var praw = [];
        var spc = "\u00a0\u00a0";
        var spc2 = spc + spc;

        // calculate all the real results
        for (var b = 0; b < benchmarks.length; b++) {
            var benchmark = benchmarks[b];
            if (pr) {
                presults.push(spc + benchmark + ":");
                praw.push(spc + benchmark + ":");
            }

            for (var m = 0; m < modes[benchmark].length; m++) {
                var mode = modes[benchmark][m];
                var bmresults = results[benchmark][mode];
                if (bmresults.length == 0) continue;

                // get the raw results
                var rr = spc2 + mode + ": [";
                for (var i = 0; i < bmresults.length; i++) {
                    if (i != 0) rr += ", ";
                    rr += bmresults[i];
                }
                rr += "]";
                if (pr) praw.push(rr);

                // now get the stats for this run
                var bmstats = stats(bmresults);

                // add it to the totals
                totals.mean += bmstats.mean;
                totals.sm += bmstats.sm;
                totals.semm += bmstats.semm;
                totals.runs++;

                // and output it
                if (pr) presults.push(spc2 + mode + ": " + bmstats.mean + "ms  " + percent(bmstats.semm) + " (stddev=" + percent(bmstats.sm) + ")");
            }

            if (pr) {
                presults.push("");
                praw.push("");
            }
        }

        // now calculate the totals
        totals.mean /= totals.runs;
        totals.sm /= totals.runs;
        totals.semm /= totals.runs;
        ptotals.push("Final results:");
        ptotals.push(spc + totals.mean + "ms  " + percent(totals.semm) + " (lower is better)");
        ptotals.push(spc + "Standard deviation = " + percent(totals.sm) + " of mean");
        if (totals.semm >= maxSEMM)
            ptotals.push(spc + "WARNING: These results are not trustworthy! After " + maxRuns + " runs, standard error divided by mean is still greater than " + percent(maxSEMM) + "!");
        else
            ptotals.push(spc + curRun + " runs");
        ptotals.push("");

        // if there are errors, mark those too
        if ("errors" in results) {
            ptotals.push("ERRORS:");
            for (var i = 0; i < results.errors.length; i++) ptotals.push(spc + results.errors[i]);
            ptotals.push("");
        }

        if (pr) {
            // and print it all out
            printarr(ptotals);
            print("Result breakdown:");
            printarr(presults);
            print("Raw results:");
            printarr(praw);
        }

        return totals;
    }

    // get statistics
    function stats(results) {
        var ret = {};

        function sum(arr) {
            return arr.reduce(function(p, c) { return p + c; }, 0);
        }

        // mean
        ret.mean = sum(results) / results.length;

        // sample stddev
        ret.stddev = Math.sqrt(
            sum(
                results.map(function(e) { return Math.pow(e - ret.mean, 2); })
            ) / (results.length - 1)
        );

        // stddev / mean
        ret.sm = ret.stddev / ret.mean;

        // sample SEM (stderr)
        ret.sem = ret.stddev / Math.sqrt(results.length);

        // sample SEM/mean
        ret.semm = ret.sem / ret.mean;

        return ret;
    }

    window.onload = onload;
})();
